import pytest
import stk

from .case_data import CaseData


@pytest.fixture(
    params=(
        lambda: CaseData(
            molecule=stk.BuildingBlock("[C+2][N+]Br"),
            canonical_atom_ids={0: 0, 1: 2, 2: 1},
        ),
        lambda: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles="Br[C+2][C+2]Br",
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit="A",
                    num_repeating_units=2,
                ),
            ),
            canonical_atom_ids={
                0: 0,
                1: 2,
                2: 4,
                3: 5,
                4: 3,
                5: 1,
            },
        ),
    ),
)
def case_data(request) -> CaseData:
    return request.param()



class CaseData:
    """
    A test case.

    Attributes
    ----------
    factory : :class:`.FunctionalGroupFactory`
        The factory being tested.

    molecule : :class:`.Molecule`
        The molecule passed to
        :meth:`.FunctionalGroupFactory.get_functional_groups`.

    functional_groups : :class:`tuple` of :class:`.FunctionalGroup`
        The expected functional groups.

    """

    def __init__(self, factory, molecule, functional_groups):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        factory : :class:`.FunctionalGroupFactory`
            The factory to test.

        molecule : :class:`.Molecule`
            The molecule, whose functional groups are found.

        functional_groups : :class:`tuple` of :class:`.FunctionalGroup`
            The expected functional groups.

        """

        self.factory = factory
        self.molecule = molecule
        self.functional_groups = functional_groups



"""
Parallel Evolutionary Algorithm
===============================

"""

import pathos

from .implementation import Implementation


class Parallel(Implementation):
    """
    A parallel implementation of the default evolutionary algorithm.

    """

    def __init__(
        self,
        initial_population,
        fitness_calculator,
        mutator,
        crosser,
        generation_selector,
        mutation_selector,
        crossover_selector,
        fitness_normalizer,
        key_maker,
        logger,
        num_processes,
    ):
        super().__init__(
            initial_population=initial_population,
            fitness_calculator=fitness_calculator,
            mutator=mutator,
            crosser=crosser,
            generation_selector=generation_selector,
            mutation_selector=mutation_selector,
            crossover_selector=crossover_selector,
            fitness_normalizer=fitness_normalizer,
            key_maker=key_maker,
            logger=logger,
        )
        self._num_processes = num_processes

    def get_generations(self, num_generations):
        with pathos.pools.ProcessPool(self._num_processes) as pool:
            yield from self._get_generations(num_generations, pool.map)



import typing
from collections.abc import Iterable
from typing import Any

from .fitness_normalizer import FitnessNormalizer

T = typing.TypeVar("T")


class NormalizerSequence(FitnessNormalizer[T]):
    """
    Applies other normalizers in sequence.

    Examples:

        *Using Multiple Fitness Normalizers*

        You want to apply multiple fitness normalizations in sequence,
        for example, by first using :class:`.DivideByMean`, followed by
        :class:`.Sum`

        .. testcode:: using-multiple-fitness-normalizers

            import stk
            import numpy as np

            building_block = stk.BuildingBlock('BrCCBr', stk.BromoFactory())
            record = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[building_block],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            fitness_values = {
                record: (1, 100, 1000),
            }
            # Create the normalizer.
            normalizer = stk.NormalizerSequence(
                fitness_normalizers=[
                    stk.DivideByMean(),
                    stk.Sum(),
                ],
            )
            normalized_fitness_values = normalizer.normalize(fitness_values)
            assert normalized_fitness_values[record] == 3
    """

    def __init__(
        self,
        fitness_normalizers: Iterable[FitnessNormalizer[T]],
    ) -> None:
        """
        Parameters:
            fitness_normalizers (list[FitnessNormalizer[T]]):
                The :class:`.FitnessNormalizer` instances which should be
                used in sequence.
        """
        self._fitness_normalizers = tuple(fitness_normalizers)

    def normalize(self, fitness_values: dict[T, Any]) -> dict[T, Any]:
        for normalizer in self._fitness_normalizers:
            fitness_values = normalizer.normalize(fitness_values)
        return fitness_values



import numpy as np
import pytest
import stk

from ...case_data import CaseData


@pytest.fixture(
    params=(
        lambda: CaseData(
            vertex=stk.metal_complex.MetalVertex(0, (1, 2, 3)),
            id=0,
            position=np.array([1, 2, 3], dtype=np.float64),
            cell=np.array([0, 0, 0]),
        ),
        lambda: CaseData(
            vertex=stk.metal_complex.MonoDentateLigandVertex(
                id=0,
                position=(1, 2, 3),
            ),
            id=0,
            position=np.array([1, 2, 3], dtype=np.float64),
            cell=np.array([0, 0, 0]),
        ),
        lambda: CaseData(
            vertex=stk.metal_complex.BiDentateLigandVertex(
                id=0,
                position=(1, 2, 3),
            ),
            id=0,
            position=np.array([1, 2, 3], dtype=np.float64),
            cell=np.array([0, 0, 0]),
        ),
    ),
)
def case_data(request) -> CaseData:
    return request.param()



from functools import partial

import stk
from scipy.spatial.distance import euclidean


def get_fg_position(id, building_block):
    (functional_group,) = building_block.get_functional_groups(id)
    return building_block.get_centroid(
        atom_ids=functional_group.get_placer_ids(),
    )


def get_centroid(building_block):
    return building_block.get_centroid()


def get_closest_point(points, point):
    return min(points, key=partial(euclidean, point))


def get_edges(vertex):
    vertex2 = stk.Vertex(1, vertex.get_position() + [-10, 0, 0])
    vertex3 = stk.Vertex(2, vertex.get_position() + [10, 0, 0])
    yield stk.Edge(0, vertex, vertex2)
    yield stk.Edge(1, vertex, vertex3)



import itertools
from functools import lru_cache

from stk._internal.json_serde.constructed_molecule import (
    ConstructedMoleculeDejsonizer,
    ConstructedMoleculeJsonizer,
)
from stk._internal.utilities.utilities import dedupe

from ..constructed_molecule import ConstructedMoleculeDatabase
from .utilities import HashableDict, get_any_value


class ConstructedMoleculeMongoDb(ConstructedMoleculeDatabase):
    """
    Uses MongoDB to store and retrieve constructed molecules.

    See Also
    --------
    :class:`.MoleculeMongoDb`
        If you need to store and retrieve molecules, which are not
        :class:`.ConstructedMolecule` instances, use a \
        :class:`.MoleculeMongoDb`.

    Examples
    --------
    *Storing and Retrieving Constructed Molecules*

    You want to store and retrieve a :class:`.ConstructedMolecule`
    from the database

    .. testsetup:: storing-and-retrieving-constructed-molecules

        import stk

        # Change the default database used, so that when a developer
        # runs the doctests locally, their "stk" database is not
        # contaminated.
        _test_database = '_stk_doctest_database'
        _old_init = stk.ConstructedMoleculeMongoDb
        stk.ConstructedMoleculeMongoDb = lambda mongo_client: (
            _old_init(
                mongo_client=mongo_client,
                database=_test_database,
            )
        )

        # Change the database MongoClient will connect to.

        import os
        import pymongo

        _mongo_client = pymongo.MongoClient
        _mongodb_uri = os.environ.get(
            'MONGODB_URI',
            'mongodb://localhost:27017/'
        )
        pymongo.MongoClient = lambda: _mongo_client(_mongodb_uri)

    .. testcode:: storing-and-retrieving-constructed-molecules

        import stk
        import pymongo

        # Connect to a MongoDB. This example connects to a local
        # MongoDB, but you can connect to a remote DB too with
        # MongoClient() - read the documentation for pymongo to see how
        # to do that.
        client = pymongo.MongoClient()
        db = stk.ConstructedMoleculeMongoDb(client)

        # Create a molecule.
        polymer = stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(
                    stk.BuildingBlock('BrCCBr', [stk.BromoFactory()]),
                ),
                repeating_unit='A',
                num_repeating_units=2,
            ),
        )

        # Place it into the database.
        db.put(polymer)

        # Retrieve it from the database.
        key_maker = stk.InchiKey()
        retrieved = db.get({
            key_maker.get_key_name(): key_maker.get_key(polymer),
        })

    .. testcode:: storing-and-retrieving-constructed-molecules
        :hide:

        _smiles = stk.Smiles()
        assert _smiles.get_key(polymer) == _smiles.get_key(retrieved)

    .. testcleanup:: storing-and-retrieving-constructed-molecules

        pymongo.MongoClient().drop_database(_test_database)
        stk.ConstructedMoleculeMongoDb = _old_init
        pymongo.MongoClient = _mongo_client

    Note that the molecule retrieved from that database can have
    a different atom ordering than the one put into it. So while the
    molecule will have the same structure, the order of the atoms
    may be different to the molecule placed into the database. This
    is because the database gives the molecules a canonical atom
    ordering, which allows position matrices to be used across
    different atom id orderings.

    *Iterating over All Entries in the Database*

    All entries in a database can be iterated over very simply

    .. testsetup:: iterating-over-all-entries-in-the-database

        import stk
        import pymongo
        import os

        # Change the database used, so that when a developer
        # runs the doctests locally, their "stk" database is not
        # contaminated.
        _test_database = '_stk_doctest_database'

        # Change the database MongoClient will connect to.

        _mongodb_uri = os.environ.get(
            'MONGODB_URI',
            'mongodb://localhost:27017/'
        )
        client = pymongo.MongoClient(_mongodb_uri)

        db = stk.ConstructedMoleculeMongoDb(
            mongo_client=client,
            database=_test_database,
         )

        # Create a molecule.
        polymer = stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(
                    stk.BuildingBlock('BrCCBr', [stk.BromoFactory()]),
                ),
                repeating_unit='A',
                num_repeating_units=2,
            ),
        )

        # Place it into the database.
        db.put(polymer)

    .. testcode:: iterating-over-all-entries-in-the-database

        for entry in db.get_all():
            # Do something to the entry.
            print(stk.Smiles().get_key(entry))

    .. testoutput:: iterating-over-all-entries-in-the-database
        :hide:

        BrCCCCBr

    .. testcleanup:: iterating-over-all-entries-in-the-database

        pymongo.MongoClient(_mongodb_uri).drop_database(_test_database)

    *Using Alternative Keys for Retrieving Molecules*

    By default, the only molecular key the database stores, is the
    InChIKey. However, additional keys can be added to the JSON stored
    in the database by using a different
    :class:`.ConstructedMoleculeJsonizer`

    .. testsetup:: using-alternative-keys-for-retrieving-molecules

        import stk

        # Change the database used, so that when a developer
        # runs the doctests locally, their "stk" database is not
        # contaminated.
        _test_database = '_stk_doctest_database'
        _old_init = stk.ConstructedMoleculeMongoDb
        stk.ConstructedMoleculeMongoDb = (
            lambda mongo_client, jsonizer: _old_init(
                mongo_client=mongo_client,
                database=_test_database,
                jsonizer=jsonizer,
            )
        )

        # Change the database MongoClient will connect to.

        import os
        import pymongo

        _mongo_client = pymongo.MongoClient
        _mongodb_uri = os.environ.get(
            'MONGODB_URI',
            'mongodb://localhost:27017/'
        )
        pymongo.MongoClient = lambda: _mongo_client(_mongodb_uri)

    .. testcode:: using-alternative-keys-for-retrieving-molecules

        import stk
        import pymongo

        db = stk.ConstructedMoleculeMongoDb(
            mongo_client=pymongo.MongoClient(),
            # Store the InChI and the InChIKey of molecules in
            # the JSON representation.
            jsonizer=stk.ConstructedMoleculeJsonizer(
                key_makers=(stk.Inchi(), stk.InchiKey()),
            ),
        )
        # Create a molecule.
        polymer = stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(
                    stk.BuildingBlock('BrCCBr', [stk.BromoFactory()]),
                ),
                repeating_unit='A',
                num_repeating_units=2,
            ),
        )
        # Places the JSON of the molecule into the database. In this
        # case, the JSON includes both the InChI and the InChIKey.
        db.put(polymer)

        # You can now use the InChI or the InChIKey to retrieve the
        # molecule from the database.
        key_maker = stk.Inchi()
        retrieved = db.get({
            key_maker.get_key_name(): key_maker.get_key(polymer),
        })

    .. testcode:: using-alternative-keys-for-retrieving-molecules
        :hide:

        _smiles = stk.Smiles()
        assert _smiles.get_key(polymer) == _smiles.get_key(retrieved)

    Obviously, most of the time, you won't have the molecule you are
    trying to retrieve from the database. Maybe you only have the
    SMILES of the molecule. You can still retrieve it

    .. testcode:: using-alternative-keys-for-retrieving-molecules

        import rdkit.Chem.AllChem as rdkit

        retrieved2 = db.get({
            'InChI': rdkit.MolToInchi(rdkit.MolFromSmiles('BrCCCCBr')),
        })

    .. testcode:: using-alternative-keys-for-retrieving-molecules
        :hide:

        assert _smiles.get_key(polymer) == _smiles.get_key(retrieved2)

    As long as you have the name of the key, and the expected value
    of the key, you can retrieve your molecule from the database.

    Note that you can create your own keys and add them to the database

    .. testcode:: using-alternative-keys-for-retrieving-molecules

        # Create your own key. This one is called "SMILES" and the
        # value is the SMILES of the molecule.
        class Smiles(stk.MoleculeKeyMaker):
            def __init__(self):
                return

            def get_key_name(self):
                return 'SMILES'

            def get_key(self, molecule):
                return rdkit.MolToSmiles(molecule.to_rdkit_mol())

        db = stk.ConstructedMoleculeMongoDb(
            mongo_client=pymongo.MongoClient(),
            jsonizer=stk.ConstructedMoleculeJsonizer(
                # Include your own custom key maker in the JSON
                # representation.
                key_makers = (stk.Inchi(), stk.InchiKey(), Smiles()),
            ),
        )

        # Place the JSON of your molecule into the database. In this
        # case the JSON will include a key called "SMILES" and
        # the value will be the SMILES of the molecule.
        db.put(polymer)

        # You can now find your molecule by using SMILES as the key,
        def normalize_smiles(smiles):
            return rdkit.MolToSmiles(
                mol=rdkit.AddHs(rdkit.MolFromSmiles(smiles)),
            )

        retrieved3 = db.get({'SMILES': normalize_smiles('BrCCCCBr')})

    .. testcode:: using-alternative-keys-for-retrieving-molecules
        :hide:

        assert _smiles.get_key(polymer) == _smiles.get_key(retrieved3)

    Often, it is unnecessary to create a whole subclass for a your
    custom key

    .. testcode:: using-alternative-keys-for-retrieving-molecules

        smiles = stk.MoleculeKeyMaker(
            key_name='SMILES',
            get_key=lambda molecule:
                rdkit.MolToSmiles(molecule.to_rdkit_mol()),
        )
        db = stk.ConstructedMoleculeMongoDb(
            mongo_client=pymongo.MongoClient(),
            jsonizer=stk.ConstructedMoleculeJsonizer(
                key_makers=(stk.InchiKey(), smiles),
            )
        )

    .. testcode:: using-alternative-keys-for-retrieving-molecules
        :hide:

        db.put(polymer)
        _retrieved4 = db.get({'SMILES': smiles.get_key(polymer)})
        assert _smiles.get_key(polymer) == _smiles.get_key(_retrieved4)

    .. testcleanup:: using-alternative-keys-for-retrieving-molecules

        pymongo.MongoClient().drop_database(_test_database)
        stk.ConstructedMoleculeMongoDb = _old_init
        pymongo.MongoClient = _mongo_client

    Note that the key you use to get the molecule back from the
    database should be unique. In other words, there should always just
    be one molecule which has that key in the database. Using a key
    that is matched by multiple molecules will likely cause your data
    to be jumbled. For example, you might return the atoms of one
    of the molecules matched by the key but holding the position matrix
    of the second molecule matched by the key.

    """

    def __init__(
        self,
        mongo_client,
        database="stk",
        molecule_collection="molecules",
        constructed_molecule_collection="constructed_molecules",
        position_matrix_collection="position_matrices",
        building_block_position_matrix_collection=(
            "building_block_position_matrices"
        ),
        jsonizer=ConstructedMoleculeJsonizer(),
        dejsonizer=ConstructedMoleculeDejsonizer(),
        put_lru_cache_size=128,
        get_lru_cache_size=128,
        indices=("InChIKey",),
    ):
        """
        Initialize a :class:`.ConstructedMoleculeMongoDb`.

        Parameters
        ----------
        mongo_client : :class:`pymongo.MongoClient`
            The database client.

        database : :class:`str`
            The name of the database to use.

        molecule_collection : :class:`str`
            The name of the collection which stores molecular
            information.

        constructed_molecule_collection : :class:`str`
            The name of the collection which stored constructed
            molecule information, that does not belong in the
            `molecule_collection`.

        position_matrix_collection : :class:`str`
            The name of the collection which stores the position
            matrices of the molecules put into and retrieved from
            the database.

        building_block_position_matrix_collection : :class:`str`
            The name of the collection, which stores the position
            matrices of the building blocks of the constructed
            molecules put into and retrieved from the database.

        jsonizer : :class:`.ConstructedMoleculeJsonizer`
            Used to create the JSON representations of molecules
            stored in the database.

        dejsonizer : :class:`.ConstructedMoleculeDejsonizer`
            Used to create :class:`.Molecule` instances from their
            JSON representations.

        put_lru_cache_size : :class:`int`, optional
            A RAM-based least recently used cache is used to avoid
            writing to the database repeatedly. This sets
            the number of values which fit into the LRU cache. If
            ``None``, the cache size will be unlimited.

        get_lru_cache_size : :class:`int`, optional
            A RAM-based least recently used cache is used to avoid
            reading from the database repeatedly. This sets
            the number of values which fit into the LRU cache. If
            ``None``, the cache size will be unlimited.

        indices : :class:`tuple` of :class:`str`, optional
            The names of molecule keys, on which an index should be
            created, in order to minimize lookup time.

        """

        database = mongo_client[database]
        self._molecules = database[molecule_collection]
        self._constructed_molecules = database[constructed_molecule_collection]
        self._position_matrices = database[position_matrix_collection]
        self._building_block_position_matrices = database[
            building_block_position_matrix_collection
        ]
        self._jsonizer = jsonizer
        self._dejsonizer = dejsonizer

        self._get = lru_cache(maxsize=get_lru_cache_size)(self._get)
        self._put = lru_cache(maxsize=put_lru_cache_size)(self._put)

        for index in indices:
            # Do not create the same index twice.
            if f"{index}_1" not in self._molecules.index_information():
                self._molecules.create_index(index)
            if (
                f"{index}_1"
                not in self._constructed_molecules.index_information()
            ):
                self._constructed_molecules.create_index(index)
            if f"{index}_1" not in self._position_matrices.index_information():
                self._position_matrices.create_index(index)

            index_information = (
                self._building_block_position_matrices.index_information()
            )
            if f"{index}_1" not in index_information:
                self._building_block_position_matrices.create_index(
                    index,
                )

    def put(self, molecule):
        molecule = molecule.with_canonical_atom_ordering()
        json = self._jsonizer.to_json(molecule)
        # lru_cache requires that the parameters to the cached function
        # are hashable objects.
        json["matrix"]["m"] = tuple(tuple(row) for row in json["matrix"]["m"])
        json["matrix"] = HashableDict(json["matrix"])
        json["molecule"] = HashableDict(json["molecule"])
        json["constructedMolecule"] = HashableDict(json["constructedMolecule"])
        json["constructedMolecule"]["BB"] = tuple(
            map(
                HashableDict,
                json["constructedMolecule"]["BB"],
            )
        )

        def make_hashable(json):
            json["matrix"]["m"] = tuple(
                tuple(row) for row in json["matrix"]["m"]
            )
            json["matrix"] = HashableDict(json["matrix"])
            json["molecule"] = HashableDict(json["molecule"])
            return HashableDict(json)

        json["buildingBlocks"] = tuple(
            map(
                make_hashable,
                json["buildingBlocks"],
            )
        )
        return self._put(HashableDict(json))

    @staticmethod
    def _get_query(json):
        keys = dict(json["matrix"])
        keys.pop("m")

        query = {"$or": []}
        for key, value in keys.items():
            query["$or"].append({key: value})
        return query

    def _put(self, json):
        query = self._get_query(json)
        self._molecules.update_many(
            filter=query,
            update={
                "$set": json["molecule"],
            },
            upsert=True,
        )
        self._position_matrices.update_many(
            filter=query,
            update={
                "$set": json["matrix"],
            },
            upsert=True,
        )

        self._add_building_block_keys_from_database(
            query=query,
            building_block_keys=json["constructedMolecule"]["BB"],
        )

        self._constructed_molecules.update_many(
            filter=query,
            update={
                "$set": json["constructedMolecule"],
            },
            upsert=True,
        )
        for building_block_json in json["buildingBlocks"]:
            building_block_query = self._get_query(building_block_json)
            self._molecules.update_many(
                filter=building_block_query,
                update={
                    "$set": building_block_json["molecule"],
                },
                upsert=True,
            )
            self._building_block_position_matrices.update_many(
                filter=building_block_query,
                update={
                    "$set": building_block_json["matrix"],
                },
                upsert=True,
            )

    def _add_building_block_keys_from_database(
        self,
        query,
        building_block_keys,
    ):
        """
        Add previously deposited keys to `building_block_keys`.

        Checks the constructed molecule collection to find all
        constructed molecule entries which match `query`. All matches
        should merely be duplicate entries for the same constructed
        molecule.

        Each entry for the constructed molecule will have a
        :class:`list` of building blocks, which were used to
        construct the constructed molecule.

        A building block is represented in this :class:`list` through
        a :class:`dict`, which maps the name of a molecular key
        (like "SMILES" or "InChIKey") to the appropriate value for that
        building block.

        The database may have multiple different dictionaries for
        the same building block, because each dictionary may hold
        different molecular keys. These differing dictionaries will be
        spread across the constructed molecule entries.

        The various different dictionaries, which all represent
        the same building block, are merged by this method. The merged
        dictionary is the one held in `building_block_keys`, which is
        updated in-place.

        This means that when `building_block_keys` is used to replace
        an entry in the database, it does not remove any building
        block keys already there.

        Parameters
        ----------
        query : :class:`dict`
            A query which matches entries, corresponding to a
            single constructed molecule.

        building_block_keys : :class:`list` of :class:`dict`
            Each :class:`dict` represents a building block of the
            constructed molecule matched by `query`. The :class:`dict`
            holds the name of a molecular key and its value for that
            particular building block. Key-value pairs for building
            block molecular keys already found in the database are
            added to the dictionaries by this method.

        Returns
        -------
        None : :class:`NoneType`

        """

        database_building_block_keys = (
            molecule_entry["BB"]
            for molecule_entry in self._constructed_molecules.find(query)
        )
        for entry_building_block_keys in database_building_block_keys:
            for keys1, keys2 in zip(
                building_block_keys,
                entry_building_block_keys,
            ):
                keys1.update(keys2)

    def get(self, key):
        # lru_cache requires that the parameters to the cached function
        # are hashable objects.
        return self._get(HashableDict(key))

    def _get(self, key):
        """
        Get the molecule with `key` from the cache.

        Parameters
        ----------
        key : :class:`.HashableDict`
            The key of a molecule, which is to be returned from the
            database.

        Returns
        -------
        :class:`.Molecule`
            The molecule held in the database under `key`.

        """

        molecule_json = self._molecules.find_one(key)
        if molecule_json is None:
            raise KeyError(
                f"No molecule found in the database with a key of: {key}"
            )

        constructed_molecule_json = self._constructed_molecules.find_one(key)
        if constructed_molecule_json is None:
            raise KeyError(
                f"No molecule found in the database with a key of: {key}"
            )

        position_matrix = self._position_matrices.find_one(key)
        if position_matrix is None:
            raise KeyError(
                "No position matrix found in the database with a key "
                f"of: {key}"
            )

        return self._dejsonizer.from_json(
            json={
                "molecule": molecule_json,
                "constructedMolecule": constructed_molecule_json,
                "matrix": position_matrix,
                "buildingBlocks": tuple(
                    map(
                        self._get_building_block,
                        constructed_molecule_json["BB"],
                    )
                ),
            },
        )

    def _get_building_block(self, key):
        return {
            "molecule": self._molecules.find_one(key),
            "matrix": self._building_block_position_matrices.find_one(key),
        }

    def get_all(self):
        # Get all potential indices.
        indices = itertools.chain(
            self._position_matrices.index_information().values(),
            self._molecules.index_information().values(),
            self._constructed_molecules.index_information().values(),
        )
        keys = tuple(
            dedupe(
                index["key"][0][0]
                for index in indices
                # Ignore "_id" index which is unique in a collection and
                # cannot be used to match molecular data split across
                # collections.
                if index["key"][0][0] != "_id"
            )
        )

        query = [
            {
                "$match": {
                    "$or": [{key: {"$exists": True}} for key in keys],
                },
            },
        ]
        query.extend(
            {
                "$lookup": {
                    "from": self._position_matrices.name,
                    "let": {
                        "molecule_key": f"${key}",
                    },
                    "as": f"posmat_{key}",
                    "pipeline": [
                        {
                            "$match": {
                                key: {"$ne": None},
                            },
                        },
                        {
                            "$match": {
                                "$expr": {
                                    "$eq": [
                                        f"${key}",
                                        "$$molecule_key",
                                    ],
                                },
                            },
                        },
                    ],
                },
            }
            for key in keys
        )
        query.extend(
            {
                "$lookup": {
                    "from": self._molecules.name,
                    "let": {
                        "molecule_key": f"${key}",
                    },
                    "as": f"mol_{key}",
                    "pipeline": [
                        {
                            "$match": {
                                key: {"$ne": None},
                            },
                        },
                        {
                            "$match": {
                                "$expr": {
                                    "$eq": [
                                        f"${key}",
                                        "$$molecule_key",
                                    ],
                                },
                            },
                        },
                    ],
                },
            }
            for key in keys
        )
        query.append(
            {
                "$match": {
                    "$expr": {
                        "$or": [
                            {
                                "$gt": [
                                    {"$size": f"$posmat_{key}"},
                                    0,
                                ],
                            }
                            for key in keys
                        ],
                    },
                },
            },
        )
        query.append(
            {
                "$match": {
                    "$expr": {
                        "$or": [
                            {
                                "$gt": [{"$size": f"$mol_{key}"}, 0],
                            }
                            for key in keys
                        ],
                    },
                },
            },
        )

        cursor = self._constructed_molecules.aggregate(query)
        for entry in cursor:
            molecule_document = get_any_value(
                mapping=entry,
                keys=(f"mol_{key}" for key in keys),
            )
            position_matrix_document = get_any_value(
                mapping=entry,
                keys=(f"posmat_{key}" for key in keys),
            )
            if (
                molecule_document is not None
                and position_matrix_document is not None
            ):
                yield self._dejsonizer.from_json(
                    {
                        "molecule": molecule_document,
                        "constructedMolecule": entry,
                        "matrix": position_matrix_document,
                        "buildingBlocks": tuple(
                            map(
                                self._get_building_block,
                                entry["BB"],
                            )
                        ),
                    }
                )



class CaseData:
    """
    A test case.

    Attributes
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    bonds : :class:`tuple` of :class:`.Bond`
        The correct bonds of :attr:`.molecule`.

    """

    def __init__(self, molecule, bonds):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        molecule : :class:`.Molecule`
            The molecule to test.

        bonds : :class:`tuple` of :class:`.Bond`
            The correct bonds of `molecule`.

        """

        self.molecule = molecule
        self.bonds = bonds



import stk

from ..utilities import is_equivalent_generic_functional_group


def test_repr(generic_functional_group):
    """
    Test :meth:`.GenericFunctionalGroup.__repr__`.

    Parameters
    ----------
    generic_functional_group : :class:`.GenericFunctionalGroup`
        The functional group, whose representation should be tested.

    Returns
    -------
    None : :class:`NoneType`

    """

    other = eval(repr(generic_functional_group), dict(stk.__dict__))
    is_equivalent_generic_functional_group(
        functional_group1=generic_functional_group,
        functional_group2=other,
    )



import pytest

import stk

from ....case_data import CaseData
from ...building_blocks import get_fe_atom, get_iron_bi_1


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                stk.metal_complex.OctahedralDelta(
                    metals={get_fe_atom(): 0},
                    ligands={get_iron_bi_1(): (0, 1, 2)},
                    reaction_factory=stk.DativeReactionFactory(
                        stk.GenericReactionFactory(
                            bond_orders={
                                frozenset(
                                    {
                                        stk.GenericFunctionalGroup,
                                        stk.SingleAtom,
                                    }
                                ): 9
                            }
                        )
                    ),
                )
            ),
            smiles=(
                "[H]C1=C([H])C([H])=[N]2->[Fe+2]34(<-[N](Br)=C([H])C2=C1[H])"
                "(<-[N](Br)=C([H])C1=C([H])C([H])=C([H])C([H])=[N]->31)<-[N]"
                "(Br)=C([H])C1=C([H])C([H])=C([H])C([H])=[N]->41"
            ),
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                stk.metal_complex.OctahedralDelta(
                    metals=get_fe_atom(),
                    ligands=get_iron_bi_1(),
                    reaction_factory=stk.DativeReactionFactory(
                        stk.GenericReactionFactory(
                            bond_orders={
                                frozenset(
                                    {
                                        stk.GenericFunctionalGroup,
                                        stk.SingleAtom,
                                    }
                                ): 9
                            }
                        )
                    ),
                )
            ),
            smiles=(
                "[H]C1=C([H])C([H])=[N]2->[Fe+2]34(<-[N](Br)=C([H])C2=C1[H])"
                "(<-[N](Br)=C([H])C1=C([H])C([H])=C([H])C([H])=[N]->31)<-[N]"
                "(Br)=C([H])C1=C([H])C([H])=C([H])C([H])=[N]->41"
            ),
            name=name,
        ),
    ),
)
def metal_complex_octahedral_delta(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



def test_with_atoms(bond, get_atom_map):
    """
    Test :meth:`.Bond.with_atoms`.

    Parameters
    ----------
    bond : :class:`.Bond`
        The bond to test.

    get_atom_map : :class:`callable`
        Takes a single parameter, `bond`, and returns a valid
        `atom_map` parameter for its :meth:`.Bond.with_atoms` method.
        This allows the testing of different values of this parameter.

    Returns
    -------
    None : :class:`NoneType`

    """

    atom_map = get_atom_map(bond)
    clone = bond.with_atoms(atom_map)
    assert clone is not bond

    expected_atom1 = atom_map.get(
        bond.get_atom1().get_id(),
        bond.get_atom1(),
    )
    assert clone.get_atom1() is expected_atom1

    expected_atom2 = atom_map.get(
        bond.get_atom2().get_id(),
        bond.get_atom2(),
    )
    assert clone.get_atom2() is expected_atom2

    assert bond.get_periodicity() == clone.get_periodicity()
    assert bond.get_order() == clone.get_order()



"""
Dative Reaction
===============

"""

from stk._internal.bond import Bond

from ..reaction.reaction import Reaction
from .utilities import is_metal


class DativeReaction(Reaction):
    """
    A reaction between two functional groups.

    The reaction creates a dative bond between the *bonder* atoms, and
    deletes any *deleter* atoms. Importantly, the direction of the bond
    is set to run from the non-metal atom to the metal atom.

    """

    def __init__(self, reaction):
        """
        Initialize a :class:`.DativeReaction` instance.

        Parameters
        ----------
        reaction : :class:`.Reaction`
            A reaction which should be made dative.

        """

        self._reaction = reaction

    def _get_new_atoms(self):
        return self._reaction._get_new_atoms()

    def _get_new_bonds(self):
        for bond in self._reaction._get_new_bonds():
            if bond.get_order() == 9 and is_metal(bond.get_atom1()):
                yield Bond(
                    atom1=bond.get_atom2(),
                    atom2=bond.get_atom1(),
                    order=bond.get_order(),
                    periodicity=bond.get_periodicity(),
                )
            else:
                yield bond

    def _get_deleted_atoms(self):
        return self._reaction._get_deleted_atoms()

    def _get_deleted_bonds(self):
        return self._reaction._get_deleted_bonds()



import stk

from .case_data import CaseData
import pathlib


def test_write(case_data: CaseData, tmp_path: pathlib.Path) -> None:
    """
    Test writing of molecule to a file.

    Parameters:
        case_data:
            A test case.

        tmp_path:
            Path to temporary directory.

    """
    _test_write(
        molecule=case_data.molecule,
        writer=case_data.writer,
        string=case_data.string,
        periodic_info=case_data.periodic_info,
        file_path=tmp_path / "tmp.coord",
    )


def _test_write(
    molecule: stk.Molecule,
    writer: stk.TurbomoleWriter,
    string: str,
    file_path: pathlib.Path,
    periodic_info: stk.PeriodicInfo | None = None,
):
    """
    Test that the written file content matches expected string.

    Parameters:
        molecule:
            Molecule to test.

        writer:
            The writer to test.

        string:
            The expected output string.

        file_path:
            Path to temporary file.

        periodic_info:
            Periodic information.

    """

    writer.write(
        molecule=molecule,
        path=file_path,
        periodic_info=periodic_info,
    )

    with open(file_path, "r") as f:
        content = f.read()

    assert content == string



"""
New Atom
========

"""

import numpy as np


class NewAtom:
    """
    Represents a new atom added by a :class:`.Reaction`.

    Examples
    --------
    *Unpacking the Atom and Position*

    The class can be elegantly unpacked into the atom and its position.

    .. testcode:: unpacking-the-atom-and-position

        import stk
        import numpy as np

        new_atoms = (
            stk.NewAtom(stk.C(-1), np.array([1., 2., 3.])),
            stk.NewAtom(stk.H(-2), np.array([2., 5., 7.])),
            stk.NewAtom(stk.F(-3), np.array([8., 2., 3.])),
        )
        for atom, position in new_atoms:
            # Do stuff.
            print(atom, position)

    .. testoutput:: unpacking-the-atom-and-position

        C(-1) [1. 2. 3.]
        H(-2) [2. 5. 7.]
        F(-3) [8. 2. 3.]

    """

    __slots__ = ["_atom", "_position"]

    def __init__(self, atom, position):
        """
        Initialize a :class:`.NewAtom` instance.

        Parameters
        ----------
        atom : :class:`.Atom`
            The atom added by the reaction. Must have a negative id.

        position : :class:`numpy.ndarray`
            The position of the new atom.

        """

        self._atom = atom
        self._position = np.array(position, dtype=np.float64)
        self._position.setflags(write=False)

    def get_atom(self):
        """
        Get the new atom.

        The new atom will have a negative id, and should be assigned
        a new one by the construction process.

        Returns
        -------
        :class:`.Atom`
            The new atom.

        """

        return self._atom

    def get_position(self):
        """
        Get the position of the new atom.

        Returns
        -------
        :class:`numpy.ndarray`
            The position of the new atom.

        """

        return self._position

    def __iter__(self):
        """
        Iterate over the atom and position.

        """

        yield self._atom
        yield self._position



import numpy as np
import pytest
import stk

from ....case_data import CaseData


@pytest.fixture(
    params=(
        lambda: CaseData(
            vertex=stk.rotaxane.AxleVertex(0, (1, 2, 3)),
            edges=(),
            building_block=stk.BuildingBlock("BrCCBr"),
            position=np.array([1, 2, 3], dtype=np.float64),
            alignment_tests={},
            functional_group_edges={},
        ),
    ),
)
def axle(request) -> CaseData:
    return request.param()



"""
Twenty Plus Thirty
==================

"""

import numpy as np

from stk._internal.topology_graphs.edge import Edge

from .cage import Cage
from .vertices import LinearVertex, NonLinearVertex


class TwentyPlusThirty(Cage):
    """
    Represents a cage topology graph.

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='BrCCBr',
            functional_groups=[stk.BromoFactory()],
        )
        bb2 = stk.BuildingBlock(
            smiles='Brc1cc(Br)cc(Br)c1',
            functional_groups=[stk.BromoFactory()],
        )
        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.TwentyPlusThirty((bb1, bb2)),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in cage.get_bonds()
            ),
        )

    Nonlinear building blocks with three functional groups are
    required for this topology.

    Linear building blocks with two functional groups are required for
    this topology.

    When using a :class:`dict` for the `building_blocks` parameter,
    as in :ref:`cage-topology-graph-examples`:
    *Multi-Building Block Cage Construction*, a
    :class:`.BuildingBlock`, with the following number of functional
    groups, needs to be assigned to each of the following vertex ids:

        | 3-functional groups: 0 to 19
        | 2-functional groups: 20 to 49

    See :class:`.Cage` for more details and examples.

    """

    # Source: http://tinyurl.com/h2dl949
    _phi = (1 + np.sqrt(5)) / 2
    _x = 1.5
    _non_linears = (
        NonLinearVertex(0, np.array([_x * _phi, 0.0, _x / _phi])),
        NonLinearVertex(1, np.array([_x * -_phi, 0.0, _x / _phi])),
        NonLinearVertex(2, np.array([_x * -_phi, 0.0, _x / -_phi])),
        NonLinearVertex(3, np.array([_x * _phi, 0.0, _x / -_phi])),
        NonLinearVertex(4, np.array([_x / _phi, _x * _phi, 0.0])),
        NonLinearVertex(5, np.array([_x / _phi, _x * -_phi, 0.0])),
        NonLinearVertex(6, np.array([_x / -_phi, _x * -_phi, 0.0])),
        NonLinearVertex(7, np.array([_x / -_phi, _x * _phi, 0.0])),
        NonLinearVertex(8, np.array([0.0, _x / _phi, _x * _phi])),
        NonLinearVertex(9, np.array([0.0, _x / _phi, _x * -_phi])),
        NonLinearVertex(10, np.array([0.0, _x / -_phi, _x * -_phi])),
        NonLinearVertex(11, np.array([0.0, _x / -_phi, _x * _phi])),
        NonLinearVertex(12, np.array([_x, _x, _x])),
        NonLinearVertex(13, np.array([_x, -_x, _x])),
        NonLinearVertex(14, np.array([-_x, -_x, _x])),
        NonLinearVertex(15, np.array([-_x, _x, _x])),
        NonLinearVertex(16, np.array([-_x, _x, -_x])),
        NonLinearVertex(17, np.array([_x, _x, -_x])),
        NonLinearVertex(18, np.array([_x, -_x, -_x])),
        NonLinearVertex(19, np.array([-_x, -_x, -_x])),
    )

    _vertex_prototypes = (
        *_non_linears,
        LinearVertex.init_at_center(
            id=20,
            vertices=(_non_linears[0], _non_linears[13]),
        ),
        LinearVertex.init_at_center(
            id=21,
            vertices=(_non_linears[0], _non_linears[12]),
        ),
        LinearVertex.init_at_center(
            id=22,
            vertices=(_non_linears[0], _non_linears[3]),
        ),
        LinearVertex.init_at_center(
            id=23,
            vertices=(_non_linears[1], _non_linears[14]),
        ),
        LinearVertex.init_at_center(
            id=24,
            vertices=(_non_linears[1], _non_linears[15]),
        ),
        LinearVertex.init_at_center(
            id=25,
            vertices=(_non_linears[1], _non_linears[2]),
        ),
        LinearVertex.init_at_center(
            id=26,
            vertices=(_non_linears[2], _non_linears[19]),
        ),
        LinearVertex.init_at_center(
            id=27,
            vertices=(_non_linears[2], _non_linears[16]),
        ),
        LinearVertex.init_at_center(
            id=28,
            vertices=(_non_linears[3], _non_linears[18]),
        ),
        LinearVertex.init_at_center(
            id=29,
            vertices=(_non_linears[3], _non_linears[17]),
        ),
        LinearVertex.init_at_center(
            id=30,
            vertices=(_non_linears[4], _non_linears[12]),
        ),
        LinearVertex.init_at_center(
            id=31,
            vertices=(_non_linears[4], _non_linears[7]),
        ),
        LinearVertex.init_at_center(
            id=32,
            vertices=(_non_linears[4], _non_linears[17]),
        ),
        LinearVertex.init_at_center(
            id=33,
            vertices=(_non_linears[5], _non_linears[6]),
        ),
        LinearVertex.init_at_center(
            id=34,
            vertices=(_non_linears[5], _non_linears[18]),
        ),
        LinearVertex.init_at_center(
            id=35,
            vertices=(_non_linears[5], _non_linears[13]),
        ),
        LinearVertex.init_at_center(
            id=36,
            vertices=(_non_linears[6], _non_linears[14]),
        ),
        LinearVertex.init_at_center(
            id=37,
            vertices=(_non_linears[6], _non_linears[19]),
        ),
        LinearVertex.init_at_center(
            id=38,
            vertices=(_non_linears[7], _non_linears[15]),
        ),
        LinearVertex.init_at_center(
            id=39,
            vertices=(_non_linears[7], _non_linears[16]),
        ),
        LinearVertex.init_at_center(
            id=40,
            vertices=(_non_linears[8], _non_linears[11]),
        ),
        LinearVertex.init_at_center(
            id=41,
            vertices=(_non_linears[8], _non_linears[12]),
        ),
        LinearVertex.init_at_center(
            id=42,
            vertices=(_non_linears[8], _non_linears[15]),
        ),
        LinearVertex.init_at_center(
            id=43,
            vertices=(_non_linears[9], _non_linears[10]),
        ),
        LinearVertex.init_at_center(
            id=44,
            vertices=(_non_linears[9], _non_linears[17]),
        ),
        LinearVertex.init_at_center(
            id=45,
            vertices=(_non_linears[9], _non_linears[16]),
        ),
        LinearVertex.init_at_center(
            id=46,
            vertices=(_non_linears[10], _non_linears[18]),
        ),
        LinearVertex.init_at_center(
            id=47,
            vertices=(_non_linears[10], _non_linears[19]),
        ),
        LinearVertex.init_at_center(
            id=48,
            vertices=(_non_linears[11], _non_linears[14]),
        ),
        LinearVertex.init_at_center(
            id=49, vertices=(_non_linears[11], _non_linears[13])
        ),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[20], _vertex_prototypes[0]),
        Edge(1, _vertex_prototypes[20], _vertex_prototypes[13]),
        Edge(2, _vertex_prototypes[21], _vertex_prototypes[0]),
        Edge(3, _vertex_prototypes[21], _vertex_prototypes[12]),
        Edge(4, _vertex_prototypes[22], _vertex_prototypes[0]),
        Edge(5, _vertex_prototypes[22], _vertex_prototypes[3]),
        Edge(6, _vertex_prototypes[23], _vertex_prototypes[1]),
        Edge(7, _vertex_prototypes[23], _vertex_prototypes[14]),
        Edge(8, _vertex_prototypes[24], _vertex_prototypes[1]),
        Edge(9, _vertex_prototypes[24], _vertex_prototypes[15]),
        Edge(10, _vertex_prototypes[25], _vertex_prototypes[1]),
        Edge(11, _vertex_prototypes[25], _vertex_prototypes[2]),
        Edge(12, _vertex_prototypes[26], _vertex_prototypes[2]),
        Edge(13, _vertex_prototypes[26], _vertex_prototypes[19]),
        Edge(14, _vertex_prototypes[27], _vertex_prototypes[2]),
        Edge(15, _vertex_prototypes[27], _vertex_prototypes[16]),
        Edge(16, _vertex_prototypes[28], _vertex_prototypes[3]),
        Edge(17, _vertex_prototypes[28], _vertex_prototypes[18]),
        Edge(18, _vertex_prototypes[29], _vertex_prototypes[3]),
        Edge(19, _vertex_prototypes[29], _vertex_prototypes[17]),
        Edge(20, _vertex_prototypes[30], _vertex_prototypes[4]),
        Edge(21, _vertex_prototypes[30], _vertex_prototypes[12]),
        Edge(22, _vertex_prototypes[31], _vertex_prototypes[4]),
        Edge(23, _vertex_prototypes[31], _vertex_prototypes[7]),
        Edge(24, _vertex_prototypes[32], _vertex_prototypes[4]),
        Edge(25, _vertex_prototypes[32], _vertex_prototypes[17]),
        Edge(26, _vertex_prototypes[33], _vertex_prototypes[5]),
        Edge(27, _vertex_prototypes[33], _vertex_prototypes[6]),
        Edge(28, _vertex_prototypes[34], _vertex_prototypes[5]),
        Edge(29, _vertex_prototypes[34], _vertex_prototypes[18]),
        Edge(30, _vertex_prototypes[35], _vertex_prototypes[5]),
        Edge(31, _vertex_prototypes[35], _vertex_prototypes[13]),
        Edge(32, _vertex_prototypes[36], _vertex_prototypes[6]),
        Edge(33, _vertex_prototypes[36], _vertex_prototypes[14]),
        Edge(34, _vertex_prototypes[37], _vertex_prototypes[6]),
        Edge(35, _vertex_prototypes[37], _vertex_prototypes[19]),
        Edge(36, _vertex_prototypes[38], _vertex_prototypes[7]),
        Edge(37, _vertex_prototypes[38], _vertex_prototypes[15]),
        Edge(38, _vertex_prototypes[39], _vertex_prototypes[7]),
        Edge(39, _vertex_prototypes[39], _vertex_prototypes[16]),
        Edge(40, _vertex_prototypes[40], _vertex_prototypes[8]),
        Edge(41, _vertex_prototypes[40], _vertex_prototypes[11]),
        Edge(42, _vertex_prototypes[41], _vertex_prototypes[8]),
        Edge(43, _vertex_prototypes[41], _vertex_prototypes[12]),
        Edge(44, _vertex_prototypes[42], _vertex_prototypes[8]),
        Edge(45, _vertex_prototypes[42], _vertex_prototypes[15]),
        Edge(46, _vertex_prototypes[43], _vertex_prototypes[9]),
        Edge(47, _vertex_prototypes[43], _vertex_prototypes[10]),
        Edge(48, _vertex_prototypes[44], _vertex_prototypes[9]),
        Edge(49, _vertex_prototypes[44], _vertex_prototypes[17]),
        Edge(50, _vertex_prototypes[45], _vertex_prototypes[9]),
        Edge(51, _vertex_prototypes[45], _vertex_prototypes[16]),
        Edge(52, _vertex_prototypes[46], _vertex_prototypes[10]),
        Edge(53, _vertex_prototypes[46], _vertex_prototypes[18]),
        Edge(54, _vertex_prototypes[47], _vertex_prototypes[10]),
        Edge(55, _vertex_prototypes[47], _vertex_prototypes[19]),
        Edge(56, _vertex_prototypes[48], _vertex_prototypes[11]),
        Edge(57, _vertex_prototypes[48], _vertex_prototypes[14]),
        Edge(58, _vertex_prototypes[49], _vertex_prototypes[11]),
        Edge(59, _vertex_prototypes[49], _vertex_prototypes[13]),
    )

    _num_windows = 12
    _num_window_types = 1



from collections import Counter

from ...utilities import DatabaseState, get_entry


def get_database_state(database):
    """
    Get the state of a :class:`.ValueMongoDb`.

    Parameters
    ----------
    database : :class:`.ValueMongoDb`
        The database whose state is wanted.

    Returns
    -------
    :class:`.DatabaseState`
        The current state of `database`.

    """

    entries = Counter()
    entries.update(map(get_entry, database._molecules.find({})))
    entries.update(map(get_entry, database._position_matrices.find({})))
    entries.update(
        map(
            get_entry,
            database._constructed_molecules.find({}),
        )
    )
    entries.update(
        map(
            get_entry,
            database._building_block_position_matrices.find({}),
        )
    )
    return DatabaseState(entries)



import pytest
from pytest_lazyfixture import lazy_fixture

# Fixtures need to be visible for lazy_fixture() calls.
from .fixtures import *  # noqa


@pytest.fixture(
    params=(
        lazy_fixture("primary_amino"),
        lazy_fixture("secondary_amino"),
        lazy_fixture("single_atom"),
        lazy_fixture("aldehyde"),
        lazy_fixture("carboxylic_acid"),
        lazy_fixture("amide"),
        lazy_fixture("thioacid"),
        lazy_fixture("alcohol"),
        lazy_fixture("alkene"),
        lazy_fixture("alkyne"),
        lazy_fixture("boronic_acid"),
        lazy_fixture("bromo"),
        lazy_fixture("dibromo"),
        lazy_fixture("difluoro"),
        lazy_fixture("diol"),
        lazy_fixture("fluoro"),
        lazy_fixture("iodo"),
        lazy_fixture("thiol"),
    ),
)
def generic_case_data(request):
    """
    A :class:`.GenericCaseData` instance.

    """

    return request.param


@pytest.fixture(
    params=(
        lazy_fixture("generic_case_data"),
        lazy_fixture("ring_amine"),
    ),
)
def case_data(request):
    """
    A :class:`.CaseData` instance.

    """

    return request.param


@pytest.fixture(
    params=(lambda n: range(n),),
)
def get_atom_ids(request):
    """
    Yield `n` atom ids.

    Parameters
    ----------
    n : :class:`int`
        The number of atom ids to yield.

    Yields
    ------
    :class:`int`
        An atom id.

    """

    return request.param


@pytest.fixture
def functional_group(case_data):
    """
    A :class:`.FunctionalGroup` instance.

    """

    return case_data.functional_group


@pytest.fixture
def generic_functional_group(generic_case_data):
    """
    A :class:`.GenericFunctionalGroup` instance.

    """

    return generic_case_data.functional_group



from .utilities import is_clone


def test_clone(vertex):
    """
    Test :meth:`.Vertex.clone`.

    Parameters
    ----------
    vertex : :class:`.Vertex`
        The vertex to test.

    Returns
    -------
    None : :class:`NoneType`


    """

    clone = vertex.clone()
    is_clone(vertex, clone)
